/*
This file is part of libmspframework
Copyright © 2006 Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/
#include <signal.h>
#include <iostream>
#include "application.h"
#include "error.h"

using namespace std;

namespace Msp {

Poller::Slot &Application::add_pollable(Pollable *obj, short events)
{
	if(!poller_)
		poller_=new Poller;
	
	Poller::Slot &slot=poller_->add_pollable(obj, events);
	// Interrupt a possible poll in progress
#ifndef WIN32  //XXX
	pthread_kill(main_tid, SIGALRM);
#endif
	return slot;
}

EventManager::Event &Application::create_event()
{
	if(!ev_mgr_)
		ev_mgr_=new EventManager(*this);

	return ev_mgr_->create_event();
}

Application::~Application()
{
	if(poller_)
		delete poller_;
	if(ev_mgr_)
		delete ev_mgr_;
}

/**
Constructs an instance of the registered application class and runs it.  If the
application throws a UsageError, the static usage() function is called.

This function can only be called once.  The global main() function provided by
the library normally does it automatically at program startup.
*/
int Application::run(int argc, char **argv)
{
	static bool called=false;
	if(called)
	{
		cerr<<"Trying to call Application::run_app twice!\n";
		return 125;
	}
	called=true;

	if(!reg_app_)
	{
		cerr<<"Trying to run with no application class registered!\n";
		return 126;
	}

#ifndef WIN32 //XXX
	signal(SIGALRM, &sigalrm_);
#endif
	
	try
	{
		app_=reg_app_->create_app(argc, argv);
	}
	catch(const UsageError &e)
	{
		reg_app_->usage(argv[0], e.get_brief());
		return 1;
	}

	int result=app_->main();
	delete app_;
	return result;
}

/**
Prints a message describing the usage of the application.  The default version
will blame the programmer for being lazy.

@param   argv0  The value of argv[0], to be used in the message
@param   brief  Whether to print a brief or long usage message
*/
void Application::usage(const char *, bool)
{
	cerr<<"The programmer was lazy and didn't write a usage() function for this application.\n";
}

Application::Application():
	exit_code(0),
	tick_mode_(IDLE),
	poller_(0),
	ev_mgr_(0)
#ifndef WIN32
	//XXX Figure out how to get the current thread on win32
	,main_tid(pthread_self())
#endif
{ }

/**
Default main loop.  Calls tick() periodically if do_ticks is true, otherwise
just sleeps.  A custom main loop should monitor the done member variable and
return exit_code.
*/
int Application::main()
{
	done=false;
	while(!done)
	{
		if(tick_mode_==IDLE)
		{
			if(poller_)
				poller_->poll(0);
			tick();
#ifdef WIN32
			Sleep(0);
#else
			sched_yield();
#endif
		}
		else
		{
			if(poller_)
				poller_->poll(-1);
			else
			{
#ifdef WIN32
				Sleep(1);
#else
				timespec ts={1000,0};
				nanosleep(&ts, 0);
#endif
			}
			if(tick_mode_!=NONE)
				tick();
		}
	}

	return exit_code;
}

/**
Sets the specified signal to be delivered to the sighandler member function.
*/
void Application::catch_signal(int s)
{
	signal(s, &sighandler_);
}

void Application::set_tick_mode(TickMode t)
{
	tick_mode_=t;
#ifndef WIN32 //XXX
	pthread_kill(main_tid, SIGALRM);	
#endif
}

/**
Causes the application to exit gracefully with the given exit code.
*/
void Application::exit(int c)
{
	done=true;
	exit_code=c;
#ifndef WIN32 //XXX
	pthread_kill(main_tid, SIGALRM);	
#endif
}

void Application::sighandler_(int s)
{
	app_->sighandler(s);
}

Application::RegBase::RegBase()
{
	if(reg_app_)
	{
		cerr<<"Warning: registering the application twice\n";
		delete reg_app_;
	}

	reg_app_=this;
}

Application *Application::app_=0;
Application::RegBase *Application::reg_app_=0;

} // namespace Msp
